Package: GenomicFiles
Title: Distributed computing by file or by range
Version: 1.48.0
Authors@R: c(
    person("Bioconductor Package Maintainer",
      email = "maintainer@bioconductor.org", role = c("aut", "cre")
    ),
    person("Valerie", "Obenchain", role = "aut"),
    person("Michael", "Love", role = "aut"),
    person("Lori", "Shepherd", role = "aut"),
    person("Martin", "Morgan", role = "aut"),
    person("Sonali", "Kumari", role = "ctb", comment =
      "Converted 'GenomicFiles' vignettes from Sweave to RMarkdown / HTML."
    ))
Description: This package provides infrastructure for parallel
    computations distributed 'by file' or 'by range'. User defined MAPPER
    and REDUCER functions provide added flexibility for data combination
    and manipulation.
License: Artistic-2.0
Depends: BiocGenerics, BiocParallel, GenomicRanges, MatrixGenerics,
        methods, Rsamtools (>= 2.25.1), rtracklayer (>= 1.69.1),
        SummarizedExperiment (>= 1.39.1)
Imports: BiocBaseUtils, GenomeInfoDb (>= 1.45.7), GenomicAlignments (>=
        1.45.1), IRanges, S4Vectors, Seqinfo, VariantAnnotation (>=
        1.55.1)
Suggests: BiocStyle, Biostrings, deepSNV, genefilter, Homo.sapiens,
        knitr, RNAseqData.HNRNPC.bam.chr14, RUnit, snpStats
VignetteBuilder: knitr
biocViews: Genetics, Infrastructure, DataImport, Sequencing, Coverage
RoxygenNote: 6.1.0
URL: https://github.com/Bioconductor/GenomicFiles
BugReports: https://github.com/Bioconductor/GenomicFiles/issues
Video: https://www.youtube.com/watch?v=3PK_jx44QTs
Collate: 'GenomicFiles-class.R' 'VcfStack-class.R'
        'reduceByFile-methods.R' 'reduceByRange-methods.R'
        'reduceFiles.R' 'reduceRanges.R' 'reduceByYield.R'
        'pack-methods.R' 'unpack-methods.R' 'registry.R' 'zzz.R'
git_url: https://git.bioconductor.org/packages/GenomicFiles
git_branch: RELEASE_3_23
git_last_commit: ad1a7e7
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 22:16:27 UTC; biocbuild
Author: Bioconductor Package Maintainer [aut, cre],
  Valerie Obenchain [aut],
  Michael Love [aut],
  Lori Shepherd [aut],
  Martin Morgan [aut],
  Sonali Kumari [ctb] (Converted 'GenomicFiles' vignettes from Sweave to
    RMarkdown / HTML.)
Maintainer: Bioconductor Package Maintainer <maintainer@bioconductor.org>
