import(methods)
importFrom(stats, rbinom)
import(BiocGenerics)
importFrom(MatrixGenerics, rowRanges)
import(S4Vectors)
import(IRanges)
import(Seqinfo)
import(GenomeInfoDb)
import(GenomicRanges)
import(SummarizedExperiment)
import(Rsamtools)
import(VariantAnnotation)
importFrom(GenomicAlignments, summarizeOverlaps, Union)
importMethodsFrom(GenomicAlignments, summarizeOverlaps)
import(rtracklayer)
import(BiocParallel)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export S4 classes
###

exportClasses(
    GenomicFiles, 
    VcfStack, RangedVcfStack
)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export S4 methods for generics not defined in GenomicFiles
###

exportMethods(
    "[", dim, names, show, yieldSize, 'colData<-',
    countBam, scanBam, summarizeOverlaps,
    coverage, summary,
    seqinfo, 'seqinfo<-',
    rowRanges, 'rowRanges<-', assay, colData,
    vcfFields
)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export S4 generics & methods for generics defined in GenomicFiles
###

export(
    ## GenomicFiles class:
    GenomicFiles, files, 'files<-',
    reduceByFile, reduceByRange, pack, unpack,

    ## VcfStack class:
    rownames, colnames
)

exportMethods(
    ## GenomicFiles class
    GenomicFiles, files, 'files<-', 

    reduceByFile, reduceByRange, pack, unpack,

    ## VcfStack class:
    rownames, colnames
)

### - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
### Export non-generic functions
###

export(
    ## file registry:
    .fileTypeRegistry, registerFileType, findTypeRegistry, makeFileType,

    reduceFiles, reduceRanges, reduceByYield, REDUCEsampler,

    VcfStack, RangedVcfStack, readVcfStack, getVCFPath, paths1kg
)
